function love.load()
	N = 62
	--N = 126

	u, v, u_prev, v_prev, dens, dens_prev = {}, {}, {}, {}, {}, {}

	local vertices = {}

	for y=0,N+1 do
		for x=0,N+1 do
			local i = I(x, y)

			vertices[i] = {x * 8 + 4, y * 8 + 4, 0, 0, 255, 0, 0, 255}
			u[i], v[i], u_prev[i], v_prev[i] = math.random(), math.random(), 0, 0
			dens[i], dens_prev[i] = math.random() * 0.3, 0
		end
	end

	mesh = love.graphics.newMesh(vertices, nil, "points")

	dirs = {}
	mags = {}
end

function love.update(dt)
	local size = 512 / (N+2)

	dt = 0.01
	vel_step(u, v, u_prev, v_prev, 0, dt)
	dens_step(dens, dens_prev, u, v, 0, dt)

	local vertices = {}

	local mag, mmag = 0, 0

	for y=0,N+1 do
		for x=0,N+1 do

			vertices[I(x, y)] = {(x+0.5) * size, (y+0.5) * size, 0, 0, 
				dens[I(x, y)] * 255, u[I(x, y)] * .255, v[I(x, y)] * .255, 255}

			u_prev[I(x, y)] = 0
			v_prev[I(x, y)] = 0
			dens_prev[I(x, y)] = 0

			local m = math.sqrt(u[I(x, y)] * u[I(x, y)] + v[I(x, y)] * v[I(x, y)]) / math.sqrt(2)

			dirs[I(x, y)] = math.atan2(v[I(x, y)], u[I(x, y)])
			mags[I(x, y)] = m

			mag = mag + m
			mmag = math.max(mmag, m)
		end
	end
	
	print(mag / ((N+2) * N+2), mmag, "> 1?")

	mesh:setVertices(vertices)

	lx, ly = mx or 0, my or 0
	mx, my = love.mouse.getPosition()
	local dx, dy = mx - lx, my - ly

	if math.abs(dx) + math.abs(dy) > 0 then
		
		local a = math.atan2(my - ly, mx - lx)
		local x, y = math.floor(mx / size), math.floor(my / size)

		if love.mouse.isDown("l") then dens_prev[I(x, y)] = 100 end
		u_prev[I(x, y)] = dx * 20
		v_prev[I(x, y)] = dy * 20
	end
end

function love.draw()
	local size = 512 / (N+2)
	love.graphics.setLineWidth(0.5)
	love.graphics.setPointSize(size)
	love.graphics.setPointStyle("rough")
	love.graphics.draw(mesh)

	--[[
	love.graphics.setColor(255, 255, 255, 192)
	for y=0,N+1 do
		for x=0,N+1 do
			local dir, mag = dirs[I(x, y)], mags[I(x, y)]
			--mag = 1
			mag = mag * 8
			local dx, dy = math.cos(dir) * size * mag, math.sin(dir) * size * mag

			--love.graphics.line(x * 8 + 4, y * 8 + 4, x * 8 + 4 + dx, y * 8 + 4 + dy)
		end
	end]]

	love.graphics.print("Current FPS: "..tostring(love.timer.getFPS( )), 10, 10)
end

function I(x, y)
	return y * (N+2) + x + 1
end

function add_source(x, s, dt)
	for i=1,(N+2)*(N+2) do
		x[i] = clamp(x[i] + dt * s[i], -1, 1)
		--s[i] = 0
	end
end

function diffuse(type, x, x0, diff, dt)
	local a = dt * diff * N*N
	solver(type, x, x0, a, 1 + 4 * a, 20)
end

function clamp(x, a, b)
	return math.max(math.min(b, x), a)
end

function advect(type, d, d0, u, v, dt)
	local dt0 = dt * N

	for j=1,N do
		for i=1,N do
			local x = i - dt0 * u[I(i, j)]
			local y = j - dt0 * v[I(i, j)]

			x = clamp(x, 0.5, N + 0.5)
			y = clamp(y, 0.5, N + 0.5)

			local i0 = math.floor(x)
			local i1 = i0 + 1
			local j0 = math.floor(y)
			local j1 = j0 + 1

			local s1 = x - i0
			local s0 = 1 - s1
			local t1 = y - j0
			local t0 = 1 - t1

			d[I(i, j)] = s0*(t0*d0[I(i0,j0)] + t1*d0[I(i0,j1)])
			           + s1*(t0*d0[I(i1,j0)] + t1*d0[I(i1,j1)])
		end
	end

	set_bnd(type, d)
end

function project(u, v)
	h = 1 / N

	local p, div = {}, {}

	for j=1,N do
		for i=1,N do
			div[I(i,j)] = h * (u[I(i+1,   j)] - u[I(i-1,   j)]
				              +v[I(  i, j+1)] - v[I(  i, j-1)]) * -0.5
			p[I(i,j)] = 0
		end
	end
	set_bnd(0, div)
	set_bnd(0, p)

	solver(0, p, div, 1, 4, 20)

	for j=1,N do
		for i=1,N do
			u[I(i,j)] = u[I(i,j)] - 0.5 / h * (p[I(i+1,   j)] - p[I(i-1,   j)])
			v[I(i,j)] = v[I(i,j)] - 0.5 / h * (p[I(  i, j+1)] - p[I(  i, j-1)])
		end
	end
	set_bnd(1, u)
	set_bnd(2, v)
end

-- Gauss-Seidel method
function solver(type, x, x0, a, c, iterations)
	for k=1,iterations do
		for j=1,N do
			for i=1,N do
				x[I(i,j)] = (x0[I(i,j)] + a * (x[I(i-1,   j)] + x[I(i+1,   j)]
					                          +x[I(  i, j-1)] + x[I(  i, j+1)])
				            ) / c
			end
		end
		set_bnd(type, x)
	end
end

function dens_step(density, input, u, v, diff, dt)
	add_source(density, input, dt)

	local diffused = input

	diffuse(0, diffused, density, diff, dt)
	advect(0, density, diffused, u, v, dt)
end

function vel_step(u, v, u0, v0, visc, dt)
	-- accept input velocities
	add_source(u, u0, dt)
	add_source(v, v0, dt)

	-- diffuse velocities
	u0, u = u, u0
	v0, v = v, v0
	diffuse(1, u, u0, visc, dt)
	diffuse(2, v, v0, visc, dt)
	project(u, v)
	-- advect velocities
	u0, u = u, u0
	v0, v = v, v0
	advect(1, u, u0, u0, v0, dt)
	advect(2, v, v0, u0, v0, dt)
	project(u, v)
end

function set_bnd(type, x)
	for i=1,N do
		if type == 1 then
			-- reflect horizontal velocities
			x[I(  0,   i)] = -x[I(1, i)]
			x[I(N+1,   i)] = -x[I(N, i)]
		else
			-- copy
			x[I(  0,   i)] = x[I(1, i)]
			x[I(N+1,   i)] = x[I(N, i)]
		end

		if type == 2 then
			-- reflect vertical velocities
			x[I(  i,   0)] = -x[I(i, 1)]
			x[I(  i, N+1)] = -x[I(i, N)]
		else
			-- copy
			x[I(  i,   0)] = x[I(i, 1)]
			x[I(  i, N+1)] = x[I(i, N)]
		end
	end

	-- average corners
	x[I(  0,   0)] = 0.5 * (x[I(1,   0)] + x[I(  0, 1)])
	x[I(  0, N+1)] = 0.5 * (x[I(1, N+1)] + x[I(  0, N)])
	x[I(N+1,   0)] = 0.5 * (x[I(N,   0)] + x[I(N+1, 1)])
	x[I(N+1, N+1)] = 0.5 * (x[I(N, N+1)] + x[I(N+1, N)])
end
